// -*-c++-*-

/***********************************************************
 * @COPYRIGHT:Jun Masaki @iesl,ies,sit 2009 
 ***********************************************************/

//////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "learning_driver.h"

#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <boost/random.hpp>

#include <cmath>

#include <rcsc/player/player_agent.h>

//0:No debug log, 1:few, 2:middle, 3:max
#define DEBUG_LEARN 0

/*-------------------------------------------------------------------*/

///////////////////////////////////////////////////////////////////////
// 行動の選択
// 学習により行動を選択する場合のみ実行
//
bool
Learning_Driver::execute( rcsc::PlayerAgent * agent )
{
  if(method_ref.isExecutable()
     && method_ref.setAction( doBoltzmannSelection(agent) ))
    return true;
  
  return false;
}

///////////////////////////////////////////////////////////////////////
// methodの初期化処理
// 初期化後はエピソードの終了判定
// execute前に呼ばれる
//
bool
Learning_Driver::prepare( rcsc::PlayerAgent * agent )
{
  if(!method_ref.isInitialized()){
    if(!initMethod(agent))
      return false;
  }else{
    return isEpisodeEnd(agent);
  }
  return true;
}

///////////////////////////////////////////////////////////////////////
// methodの初期化処理
// 学習設定ファイルと学習データファイルを読み込む
// 学習設定ファイルがないときは、デフォルトの設定を使う
// 学習データファイルがないときは、新たに学習データファイルを作る
// 処理を行ったあとは初期化終了処理を行う
// このメソッドは起動後1度しか呼ばれない
//
bool
Learning_Driver::initMethod( rcsc::PlayerAgent * agent )
{
  //unumを学習データの識別子とする
  method_ref.setMyUnum(agent->world().self().unum());

  if(!readSettingFile()){
    method_ref.setDefaultSettings();
    method_ref.processWhenInitialized();
    return false;
  }
  method_ref.processWhenInitialized();

  if(method_ref.isExecutable()
     && !readLearningDataFile()
     && method_ref.writesData()){
    makeNewLearningDataFile();
  }
  method_ref.processWhenReadData();
 
  return true;
}

///////////////////////////////////////////////////////////////////////
// 学習設定ファイルを読む
// ファイル名、設定項目はmethodから呼び出す
// 読み込んだ設定はmethodに設定する
//
bool
Learning_Driver::readSettingFile()
{
  std::ifstream ifs(method_ref.getSettingFileName().c_str() , std::ios::in);
  if(!ifs){
    std::cout 
      << "Reading:" << method_ref.getSettingFileName() 
      << " This file could not be read."
      << std::endl;
    return false;
  }
#if DEBUG_LEARN >= 1
  std::cout << "Reading:" << method_ref.getSettingFileName() 
	    << " Open."
	    << std::endl;
#endif

  typedef boost::tokenizer< boost::char_separator<char> > tokenizerCont;
  boost::char_separator<char> separetor(":");
  std::string currentLine;
  while(getline(ifs, currentLine)){

    //行の先頭が "#" または 空行 なら読み飛ばす
    if(currentLine.substr(0,1) == "#" || currentLine.empty())
      continue;
    
    tokenizerCont tokens(currentLine , separetor);
    tokenizerCont::iterator token = tokens.begin();
    
    //設定ファイルの項目とmethodの中の設定項目と照会、設定ファイルから設定をロード
    int i = 0;
    while(method_ref.getSettingItem(i)){
      if(*token == method_ref.getSettingItem(i)->getName()){
	++token;
	if(*token == "on" || *token == "true"){
	  method_ref.getSettingItem(i)->setParam(true);
	  break;
	}else if(*token == "off" || *token == "false"){
	  method_ref.getSettingItem(i)->setParam(false);
	  break;
	}else if(boost::regex_match(*token,boost::regex("-{0,1}[0-9]+|-{0,1}[0-9]+.[0-9]+"))){
	  try{
	    method_ref.getSettingItem(i)->setParam(boost::lexical_cast<double>(*token));
	  }catch(boost::bad_lexical_cast & ex){
	    std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
	    ifs.close();
	    return false;
	  }
	  break;
	}else{
	  method_ref.getSettingItem(i)->setParam(std::string(*token));
	  break;
	}
      }
      i++;
    }
  }
  
  ifs.close();
  return true;
}

///////////////////////////////////////////////////////////////////////
// 学習データファイルを読み込む
// ファイル名、カラム名はmethodから呼び出す
// 読み込んだタプルはmethodに設定する
//
bool 
Learning_Driver::readLearningDataFile()
{
  //ファイルを読み取り専用で開く
  std::ifstream ifs(method_ref.getLearningDataFileName().c_str(), std::ios::in);
  if(!ifs){
    std::cout 
      << "Reading:" << method_ref.getLearningDataFileName() 
      << " This file could not be read."
      << std::endl;
    return false;
  }
  
#if DEBUG_LEARN >= 1
  std::cout << "Reading:" << method_ref.getLearningDataFileName() 
	    << " Open."
	    << std::endl;
#endif

  typedef boost::tokenizer< boost::char_separator<char> > tokenizerCont;
  boost::char_separator<char> separator(",");
  std::string currentLine;
  std::string lastLine;
  int i;

  //学習ファイルの項目とmethodの中の設定項目と照会
  if(getline(ifs, currentLine)){
    tokenizerCont tokens(currentLine , separator);
    tokenizerCont::iterator token_it = tokens.begin();
    i = 0;
  
    //項目名の対応関係を検証
    while(method_ref.getLearningDataItem(i)){
      if( *token_it != method_ref.getLearningDataItem(i)->getName() ){
	ifs.close();
#if DEBUG_LEARN >= 1
	std::cout << "Reading:" << method_ref.getLearningDataFileName()
		  << " A column name in the data didn't match with the method's one."
		  << std::endl;
#endif
	return false;
      }
      
      ++token_it;
      i++;
      //終了判定：method側に未対応の項目名がある場合、読み込み失敗
      if(token_it==tokens.end()){
	if(method_ref.getLearningDataItem(i)){
	  ifs.close();
#if DEBUG_LEARN >= 1
	  std::cout << "Reading:" << method_ref.getLearningDataFileName()
		    << " There are too much columns in the data."
		    << std::endl;
#endif
	  return false;
	}
	break;
      }
    }
    //学習ファイル側に未対応の項目名がある場合、読み込み失敗
    if(token_it!=tokens.end()){
      ifs.close();
#if DEBUG_LEARN >= 1
      std::cout << "Reading:" << method_ref.getLearningDataFileName()
		<< " There are not enough columns in the data."
		<< std::endl;
#endif
      return false;
    }
  }else{
    ifs.close();
#if DEBUG_LEARN >= 1
    std::cout << "Reading:" << method_ref.getLearningDataFileName()
	      << " There are not data in this file."
	      << std::endl;
#endif
    return false;
  }

  //最新の学習データ以外は読み飛ばす
  while(getline(ifs, currentLine)){
    lastLine = currentLine;
  }
  
  if(lastLine.empty()){
    ifs.close();
#if DEBUG_LEARN >= 1
    std::cout << "Reading:" << method_ref.getLearningDataFileName()
	      << " The data are not correct."
	      << std::endl;
#endif
    return false;
  }

  //学習ファイルのデータをロード
  tokenizerCont tokens(lastLine , separator);
  tokenizerCont::iterator token_it = tokens.begin();
  i = 0;
  
  while(method_ref.getLearningDataItem(i)){
    //boostが例外を投げる可能性があるので、tryで囲んでおく
    try{
      //読み飛ばさない項目で値を設定できない項目がある場合は、読み込み失敗
      if( !method_ref.getLearningDataItem(i)->passes() 
	  && !method_ref.getLearningDataItem(i)->setParam( boost::lexical_cast<double>(*token_it) ) ){
	ifs.close();
#if DEBUG_LEARN >= 1
	std::cout << "Reading:" << method_ref.getLearningDataFileName()
		  << " The data could not be setted."
		  << std::endl;
#endif
	return false;
      }
    }catch(boost::bad_lexical_cast & ex){
      std::cout << __FILE__ << ":" << __LINE__ << ":" << ex.what() << std::endl;
      ifs.close();
      return false;
    }
    
    ++token_it;
    i++;
    //終了判定：method側に未対応の項目がある場合、読み込み失敗
    if(token_it==tokens.end()){
      if(method_ref.getLearningDataItem(i)){
	ifs.close();
#if DEBUG_LEARN >= 1
	std::cout << "Reading:" << method_ref.getLearningDataFileName()
		  << " There are too much items in the data."
		  << std::endl;
#endif
	return false;
      }
      break;
    }
  }
  //学習ファイル側に未対応の項目がある場合、読み込み失敗
  if(token_it!=tokens.end()){
    ifs.close();
#if DEBUG_LEARN >= 1
      std::cout << "Reading:" << method_ref.getLearningDataFileName()
		<< " There are not enough items in the data."
		<< std::endl;
#endif
    return false;
  }

  ifs.close();
  return true;
}

///////////////////////////////////////////////////////////////////////
// 学習データファイルを新たに作る
// ファイル名、カラム名はmethodから呼び出す
//
bool
Learning_Driver::makeNewLearningDataFile(){
  //ファイルを書き込みモードで新たに作る
  std::ofstream ofs(method_ref.getLearningDataFileName().c_str()
		    ,std::ios::out);
  if( !ofs ){
    std::cout 
      << "Making:" << method_ref.getLearningDataFileName() 
      << " This file could not be opened."
      << std::endl;
    return false;
  }

#if DEBUG_LEARN >= 1
  std::cout << "Making:" << method_ref.getLearningDataFileName() 
	    << " Open."
	    << std::endl;
#endif

  for(int i=0;method_ref.getLearningDataItem(i);i++){
    ofs << method_ref.getLearningDataItem(i)->getName();
    if(method_ref.getLearningDataItem(i+1))
      ofs << ",";
    else
      ofs << std::endl;
  }

  ofs.close();
  return true;
}

///////////////////////////////////////////////////////////////////////
// エピソードの終了を判定
//
bool
Learning_Driver::isEpisodeEnd( rcsc::PlayerAgent * agent )
{
  if(method_ref.learns()
     && method_ref.isEpisodeEnd(agent)){
    return endTheEpisode(agent);
  }
  
  return false;
}

///////////////////////////////////////////////////////////////////////
// エピソードの終了処理
// 報酬を更新する
// 重みを更新する
// 学習データをファイルに書き込む
//
bool
Learning_Driver::endTheEpisode( rcsc::PlayerAgent * agent )
{
  if(!method_ref.isInitialized())
    return false;

  if( method_ref.updateRewards(agent)
      && method_ref.updateWeights(agent)
      && method_ref.endTheEpisode(agent)
      && method_ref.writesData()){
    writeLearningDataFile();
  }
  return true;
}

///////////////////////////////////////////////////////////////////////
// 学習データをファイルに書き込む
// ファイル名、書き込む値はmethodから呼び出す
//
bool
Learning_Driver::writeLearningDataFile(){
  // ファイルを追記モードで開く
  std::ofstream ofs(method_ref.getLearningDataFileName().c_str()
		    ,std::ios::out | std::ios::app);
  if( !ofs ){
    std::cout 
      << "Write:" << method_ref.getLearningDataFileName() 
      << " This file could not be opened."
      << std::endl;
    return false;
  }
  
  ofs.setf(std::ios::fixed);
  for(int i=0;method_ref.getLearningDataItem(i);i++){
    if(method_ref.getLearningDataItem(i)->hasSettedParam()
       &&method_ref.getLearningDataItem(i)->hasSettedNoBool()){
      if(!method_ref.getLearningDataItem(i)->getParam(std::string("string")).empty())
	{
	  ofs << method_ref.getLearningDataItem(i)->getParam(std::string("string"));
	}
      else
	{
	  ofs << method_ref.getLearningDataItem(i)->getParam(0.0);
	}
      if(method_ref.getLearningDataItem(i+1))
	ofs << ",";
      else
	ofs << std::endl;
    }else{
      if(!method_ref.getLearningDataItem(i+1))
	ofs << std::endl;
    }
  }
  
  ofs.close();
#if DEBUG_LEARN >= 2
  std::cout  << "Write:" << method_ref.getLearningDataFileName()
	     << " Open."
	     << std::endl;
#endif
  method_ref.processWhenSavedLearningData();
  return true;
}

///////////////////////////////////////////////////////////////////////
//ボルツマン分布により、確率的に行動を選択
//
int
Learning_Driver::doBoltzmannSelection( rcsc::PlayerAgent * agent ){
  std::vector<double> probabilities;  //各行動の確率
  std::vector<double> energies;       //各行動のエネルギー
  double denominator = 0.0;           //ボルツマン分布の分母

  //行動を生成し、エネルギーを計算。併せて、ボルツマン分布の分母を計算する。
  if(method_ref.makeAction(agent)){
    int action = 0;
    do{
      double actionEnergy = calcEnergy(action,agent);
      if(actionEnergy > 1000){
#if DEBUG_LEARN >= 2
	std::cout << agent->world().self().unum()
		  << ":[" << agent->world().time().cycle() << "]"
		  << " The energy is too high:"
		  << " action:" << action
		  << " energy:" << actionEnergy
		  << std::endl;
#endif
	return -1;
      }

      energies.push_back(actionEnergy);
      //注：e^x : x は+-700以内であるべき
      denominator += std::exp( - actionEnergy / method_ref.getTemperature());
      if(isinf(denominator)){
#if DEBUG_LEARN >= 2
	std::cout << agent->world().self().unum()
		  << ":[" << agent->world().time().cycle() << "]"
		  << " The bolzmann distribution's denominator is infinity."
		  << std::endl;
#endif
	return -1;
      }

#if DEBUG_LEARN >= 3
    std::cout << agent->world().self().unum()
	      << ":[" << agent->world().time().cycle() << "]"
	      << " each_energy:"
	      << " action:" << action
	      << " energy:" << actionEnergy
	      << " accumulate_denominator:" << denominator
	      << std::endl;
#endif

      action++;
    }while(action < method_ref.getActionSize());
  }else{
#if DEBUG_LEARN >= 2
    std::cout << agent->world().self().unum()
	      << ":[" << agent->world().time().cycle() << "]"
	      << " could not make actions."
	      << std::endl;
#endif
    return -1;
  }
 
  //各行動のエネルギーを計算
  std::vector<double>::const_iterator it = energies.begin();
  std::vector<double>::const_iterator end = energies.end();
  for(;it!=end;++it){
    double probability = 
      std::exp( - (*it) / method_ref.getTemperature() ) 
      / denominator;
    if(isnan(probability) || isinf(probability)){
#if DEBUG_LEARN >= 2
      std::cout << agent->world().self().unum()
		<< ":[" << agent->world().time().cycle() << "]"
		<< " The probability is Not a number or infinity:"
		<< " energy:" << *it
		<< " denominator:" << denominator 
		<< std::endl;
#endif
      return -1;
    }
#if DEBUG_LEARN >= 3
    std::cout << agent->world().self().unum()
	      << ":[" << agent->world().time().cycle() << "]"
	      << " each_prob.:"
	      << " prob.:" << probability
	      << std::endl;
#endif
    probabilities.push_back(probability);
    method_ref.storeProb(probability);
  }
  
  //乱数を生成
  boost::mt19937 engine(static_cast<uint32_t>(time(0)));
  boost::uniform_real<> distribution(0,1);
  boost::variate_generator< boost::mt19937 , boost::uniform_real<> > 
    rand(engine,distribution);  
  const double RANDOM = rand();

  //各行動の確率を足していき、乱数を越えた時の行動が選択された行動
  double decider = 0.0;
  it = probabilities.begin();
  end = probabilities.end();
  for(int i = 0;it != end;i++,++it){

    decider += *it;
    if(RANDOM <= decider){
      #if DEBUG_LEARN >= 2
      std::cout << agent->world().self().unum()
		<< ":[" << agent->world().time().cycle() << "]"
		<< " decide the action:" << i
		<< " its prob.:" << *it
		<< " RANDAM:" << RANDOM
		<< " sum_prob.:" << decider
		<< std::endl;
#endif
      return i;
    }
  }

#if DEBUG_LEARN >= 2
  std::cout << agent->world().self().unum()
	    << ":[" << agent->world().time().cycle() << "]"
	    << " could not decide an action:"
	    << " RANDAM:" << RANDOM
	    << " sum_prob.:" << decider
	    << std::endl;
#endif
  return -1;
}
 
///////////////////////////////////////////////////////////////////////
//行動に対するエネルギーを計算する
//
double
Learning_Driver::calcEnergy( int action , rcsc::PlayerAgent * agent )
{
  if(action == -1)
    return 10000.0;

  double energy = 0.0;
  //重みとヒューリスティックを掛け合わせて足す
  for(int i=0; i < method_ref.getHeuristicsSize(); i++)
    {
      double uValue = method_ref.U(i , action , agent);
      double weight = method_ref.getWeight(i);
#if DEBUG_LEARN >= 3
      std::cout << agent->world().self().unum()
		<< ":[" << agent->world().time().cycle() << "]"
		<< " heuristics_energy:" << weight * uValue
		<< " U(action:" << action
		<< " ,heuristics:" << i
		<< " ):" << uValue
		<< " weight:" << weight
		<< std::endl;
#endif
      energy -= weight * uValue;
    }
  return energy;
}

